// Set up relay board
#define relayLatch 12
#define relayClock 13
#define relayData 14
boolean relay_state[relay_count];
unsigned long relay_delay = 5000;
unsigned long relay_next = relay_delay;
#define relay_count 4
int relay_image[] = {2,2,1,3};

// BME 280
#include <Wire.h>
#include "Adafruit_BME280.h"
#define I2C_SDA 5
#define I2C_SCL 4
#define BME280_ADDRESS 0x76
#define atmosPressure 0.83
#define Pa2Hg 3386.39
Adafruit_BME280 bme(I2C_SDA, I2C_SCL);

// OLED SCREEN PARAMETERS
#define SCREEN_I2C_ADDRESS 0x3C
#define SCREEN_SDA_PIN 5 //SDA
#define SCREEN_SCL_PIN 4 //SCL
#define SCREEN_HEIGHT 64
#define SCREEN_WIDTH 128
#define SCREEN_VERTICAL_FLIP true

#include "font.h"
#include "icons.h"
#include <SSD1306Wire.h>
SSD1306Wire display(SCREEN_I2C_ADDRESS, SCREEN_SDA_PIN, SCREEN_SCL_PIN);

// Configure WiFi
#include <WiFi.h>
const char* SSID = "Your_SSID";
const char* PASS = "Your_PASS";
const char* HOST = "example.com";
const char* URI = "path/to/sensors.php";
#define PORT 80
int32_t rssi = 0;
unsigned long check_wifi = 30000;
WiFiClient client;

// Light sensor
#define sunlight_adc ADC1_CHANNEL_0
unsigned int sunlight = 0;

// moisture sensors
#define moisture_count 2
const adc1_channel_t moisture_adc[] = {ADC1_CHANNEL_6, ADC1_CHANNEL_7};
unsigned int moisture[] = {50,50};

// Configure global variables
unsigned long NOW = millis();
unsigned long timestamp = 0;
int timezone = 0;
unsigned long clockdiff = 0;
boolean flash = false;
unsigned long next_flash;
unsigned long next_read;
unsigned long next_show;
unsigned long next_rssi;
unsigned long send_data = 5000;
boolean got_valid_data = false;
float celcius;
float fahrenheit;
float humidity;
float pressurePa;

// Parameters for the scrolling text
#define scroll_delay 20          // set the scroll speed
#define scroll_top 53              // row of text line
#define scroll_height 10         // character height
int scroll_pos = SCREEN_WIDTH;
unsigned int next_scroll = 0;
const char* scroll_div = "  |  ";
const char *scroll_font = DejaVu_Sans_9;


void getValues(void);


void setup() {
  Serial.begin(115200);

  Wire.begin(SCREEN_SDA_PIN, SCREEN_SCL_PIN);
  Wire.beginTransmission(SCREEN_I2C_ADDRESS);
  // Check OLED is connected before configuring it
  if(Wire.endTransmission() == 0) {
    display.init();
    display.setContrast(255);
    if(SCREEN_VERTICAL_FLIP) display.flipScreenVertically();
    display.clear();
    display.setFont(ArialMT_Plain_16);
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(63, 26, "Waiting for data");
    display.display();
  } else {
    Serial.print("OLED screen not found at 0x");
    Serial.println(SCREEN_I2C_ADDRESS, HEX);
    while (1);
  }


  WiFi.enableSTA(true);
  WiFi.begin(SSID, PASS);


  if (! bme.begin(BME280_ADDRESS)) {
    Serial.println("Could not find a valid BME280 sensor!");
    while (1);
  }
  bme.setSampling(Adafruit_BME280::MODE_FORCED, 
  Adafruit_BME280::SAMPLING_X1, // temperature 
  Adafruit_BME280::SAMPLING_X1, // pressure 
  Adafruit_BME280::SAMPLING_X1, // humidity 
  Adafruit_BME280::FILTER_OFF   );


// Configure analog inputs
  adc1_config_width(ADC_WIDTH_BIT_12);
  adc1_config_channel_atten(sunlight_adc, ADC_ATTEN_DB_0);
  adc1_config_channel_atten(moisture_adc[0], ADC_ATTEN_DB_0);
  adc1_config_channel_atten(moisture_adc[1], ADC_ATTEN_DB_0);


  // Configure relays
  pinMode(relayLatch, OUTPUT);
  pinMode(relayClock, OUTPUT);
  pinMode(relayData, OUTPUT);
  for (int i=0; i<16; i++) relay_state[i] = (i < relay_count) ? 0 : 1;
  setRelays();
}


void loop() {
  NOW = millis();

  /***** Get input from website *****/
  char buffer[12];
  if(client.available()) {
    // only read one line at a time so as not to block other functions
    String line = client.readStringUntil('\n');

    if (line.indexOf("TIMEZONE:") == 0) {
      timezone = line.substring(9).toInt();
    }

    if (line.indexOf("TIMESTAMP:") == 0) {
      unsigned long tm = line.substring(10).toInt();
      timestamp = tm;
      clockdiff = tm - (NOW / 1000);
    }
  }

  /***** check timing on functions *****/
  if (NOW >= next_scroll) {
    scroll_pos--;
    next_scroll = NOW + scroll_delay;
    showScroll();
  }

  if (NOW >= next_flash) {
    flash = !flash;
    next_flash = NOW + 500;
  }

  if (NOW >= next_rssi) {
    //rssi = getRSSI(SSID);
    rssi = WiFi.RSSI();
    next_rssi = NOW + 1000;
  }

  if (NOW >= next_read) {
    Serial.println("...Read values...");
    getValues();
    next_read = NOW + 60000;
  }

  if (NOW >= next_show) {
    showValues();
    next_show = NOW + 250;
  }

  if (NOW >= send_data) {
    Serial.println("...Send values...");
    sendValues();
    send_data = NOW + 30000;
  }

  // if wifi is down, try reconnecting every 30 seconds
  if ((WiFi.status() != WL_CONNECTED) && (NOW > check_wifi)) {
    Serial.println("...Reconnecting WiFi...");
    WiFi.disconnect();
    WiFi.begin(SSID, PASS);
    check_wifi = NOW + 30000;
  }

  if (NOW >= relay_next) {
    unsigned long clocktime = time();
    int h = ((clocktime / 3600) + timezone) % 24;
    int m = (clocktime / 60) % 60;
/*
    // randomize relay settings
    if (! random(10)) {
      int rnd = random(relay_count);
      relay_state[rnd] = (random(4)) ? false : true;
    }
*/
    relay_state[0] = (fahrenheit < 75) ? true : false; // 55
    relay_state[1] = (fahrenheit < 72) ? true : false; // 45
    relay_state[2] = ((m/10) % 2) ? true : false;
    relay_state[3] = ((sunlight < 25) && (h>=6 && h<=20)) ? true : false;

    setRelays();
    relay_next = NOW + relay_delay;
  }

  delay(10);
}


unsigned long time() {
  return round(clockdiff + (millis() / 1000));
}


void getValues() {
  bme.takeForcedMeasurement();

  float C = bme.readTemperature();
  //Serial.println(C);

  // Range: -144.68 to 193?
  if ((C > -144) && (C < 193)) {
    //Serial.println("Values OK");
    celcius = C;
    fahrenheit = (celcius * 1.8) + 32;  // F = round(F * 10) / 10;
    humidity = bme.readHumidity();
    pressurePa = bme.readPressure();
    got_valid_data = true;
  } else got_valid_data = false;
}


void showValues() {
  if (! got_valid_data) return;
  String txt;

  display.setColor(BLACK);
  display.fillRect(0,0, SCREEN_WIDTH,53);
  display.setColor(WHITE);
  display.setTextAlignment(TEXT_ALIGN_RIGHT);

  /***** display temperature and humidity *****/
  // frame around temp/humidity
  display.drawLine(4, 26, 58, 26);
  display.drawLine(2,0,60,0);
  display.drawLine(60,0,62,2);
  display.drawLine(62,2,62,50);
  display.drawLine(62,50,60,52);
  display.drawLine(60,52,2,52);
  display.drawLine(2,52,0,50);
  display.drawLine(0,50,0,2);
  display.drawLine(0,2,2,0);

  // temperature
  if (fahrenheit >= 100) txt = String(fahrenheit, 0);
    else txt = String(fahrenheit, 1);
    txt += "°";
  display.setFont(ArialMT_Plain_24);
  display.drawString(60, 0, txt);

  // humidity
  txt = String(int(humidity));
  display.setFont(ArialMT_Plain_24);
  display.drawString(44, 26, txt);
  display.setFont(ArialMT_Plain_16);
  display.drawString(58, 28, "%");

  /***** draw bar graphs *****/
  // wifi strength
  display.drawFastImage(66,0, 9,7, icoWifi);
  display.drawProgressBar(77,0, 50,6, abs(rssi));

  // sunlight
  sunlight = adc1_get_raw(sunlight_adc) * (100.0/4095.0);
  display.drawFastImage(66,9, 9,7, icoSun);
  display.drawProgressBar(77,9, 50,6, sunlight);

  // soil moisture
  for(int i=0; i<moisture_count; i++) {
    moisture[i] = adc1_get_raw(moisture_adc[i]) * (100.0/4095.0);
  }
  display.drawFastImage(64,20, 12,12, icoLgWater);
  display.drawProgressBar(77,18, 50,6, moisture[0]);
  display.drawProgressBar(77,27, 50,6, moisture[1]);

  /***** display relays *****/
  // left brace
  display.setPixel(67,38);
  display.setPixel(66,39);
  display.drawLine(65,40,65,50);
  display.setPixel(66,51);
  display.setPixel(67,52);

  // right brace
  display.setPixel(125,38);
  display.setPixel(126,39);
  display.drawLine(127,40,127,50);
  display.setPixel(126,51);
  display.setPixel(125,52);

  // Relay status
  for(int i=0; i<int(relay_count/8+1)*8; i++) {
    int x = (7 * i); x += (i<8) ? 69 : 13;
    int y = (i<8) ? 38 : 46;

    int img = 0;
    if (i < relay_count) {
      img = relay_image[i];
      if (! relay_state[i]) img = 0;
      if (! flash) img = 0;
    } else img = 99;

    // 0=dot, 1=fan, 2=heat, 3=light, 4=water, 5=check, 6=lock
    switch(img) {
      case 1: display.drawFastImage(x,y, 6,7, icoFan); break;
      case 2: display.drawFastImage(x,y, 6,7, icoHeat); break;
      case 3: display.drawFastImage(x,y, 6,7, icoLight); break;
      case 4: display.drawFastImage(x,y, 6,7, icoWater); break;
      case 5: display.drawFastImage(x,y, 6,7, icoCheck); break;
      case 6: display.drawFastImage(x,y, 6,7, icoLock); break;
      case 99: display.drawFastImage(x,y, 6,7, icoNone); break;

      default: display.drawFastImage(x,y, 6,7, icoOff); break;
    }
  }

  display.display();
}


void showScroll() {
  String txt = "";
  txt += String(round(pressurePa / atmosPressure / 100.0F)) + " hPa" + String(scroll_div);
  txt += String(pressurePa / Pa2Hg / atmosPressure) + " Hg" + String(scroll_div);
  txt += WiFi.localIP().toString() + String(scroll_div);
  if (clockdiff) {
    unsigned long clocktime = time();
    int h = ((clocktime / 3600) + timezone) % 24;
    String ap = "am";
    if (h > 12) { h -= 12; ap = "pm"; }
    if (h == 0) h = 12;
    int m = (clocktime / 60) % 60;
    int s = clocktime % 60;
    char buff[9]="00:00:00";
    sprintf(buff, "%d:%02d:%02d", h, m, s);

    txt += String(buff) + ap + String(scroll_div);
  }

  uint16_t firstChar = pgm_read_byte(scroll_font + FIRST_CHAR_POS);
  uint16_t strWidth = 0;
  for (int i=0; i<txt.length(); i++) {
    strWidth += pgm_read_byte(scroll_font + JUMPTABLE_START + (txt[i] - firstChar) * JUMPTABLE_BYTES + JUMPTABLE_WIDTH);
  }
  yield();

  if (scroll_pos + strWidth <= SCREEN_WIDTH) scroll_pos += strWidth;

  display.setColor(BLACK);
  display.fillRect(0, scroll_top+1, SCREEN_WIDTH, scroll_height);
  display.setColor(WHITE);

  display.setFont(scroll_font);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(scroll_pos, scroll_top, txt);
  int pos = (scroll_pos > 0) ? (scroll_pos - strWidth) : (scroll_pos + strWidth);
  display.drawString(pos, scroll_top, txt);

  display.display();
}


void sendValues() {
  if (! got_valid_data) return;
  if (WiFi.status() != WL_CONNECTED) return;

  String txt;
  txt  = "{";
  if (clockdiff) {
    txt += "\"timestamp\":";
    txt += time();
    txt += ",";
  }
  txt += "\"temperature\":";
  txt += fahrenheit;
  txt += ",\"humidity\":";
  txt += humidity;
  txt += ",\"pressure\":";
  txt += pressurePa;

  txt += ",\"sunlight\":";
  txt += sunlight;

  txt += ",\"moisture\":[";
  for (int i=0; i < moisture_count; i++) {
    txt += moisture[i];
    if (i < (moisture_count - 1)) txt += ",";
  }
  txt += "]";

  txt += ",\"relay\":[";
  for (int i=0; i < relay_count; i++) {
    txt += relay_state[i];
    if (i < (relay_count - 1)) txt += ",";
  }
  txt += "]";

  txt += "}";
  Serial.println(txt);


  if (!client.connected()) {
    Serial.println("*** Connecting to server...");
    if (!client.connect(HOST, PORT)) {
      Serial.println("connection failed");
      // Store current data in buffer until connection resumes
      return;
    }
  }

  Serial.println("Connected");
  client.print("GET " + String(URI) + "?data=" + txt);
  client.print(" HTTP/1.1\r\n");
  client.print("Host: " + String(HOST) + "\r\n");
  client.print("Connection: close\r\n\r\n");
}


void setRelays() {
  unsigned long result = 0;
  for (int i=0; i<relay_count; i++) {
    if (relay_state[i]) result = result | (1 << i);
  }

  digitalWrite(relayLatch, LOW);
  shiftOut(relayData, relayClock, MSBFIRST, result);
  digitalWrite(relayLatch, HIGH);
}

